from plivo.utils.validators import *
from plivo.xml import PlivoXMLElement, map_type


class MultiPartyCallElement(PlivoXMLElement):
    _name = 'MultiPartyCall'
    _nestable = []

    @property
    def max_duration(self):
        return self.__max_duration

    @max_duration.setter
    @validate_args(
        max_duration=[
            optional(
                of_type_exact(int),
                check(lambda max_duration: 300 <= max_duration <= 28800, '300 <= max_duration <= 28800')
            )]
    )
    def max_duration(self, max_duration):
        self.__max_duration = max_duration

    def set_max_duration(self, max_duration):
        self.max_duration = max_duration
        return self

    @property
    def max_participants(self):
        return self.__max_participants

    @max_participants.setter
    @validate_args(
        max_participants=[
            optional(
                of_type_exact(int),
                check(lambda max_participants: 2 <= max_participants <= 10, '2 <= max_participants <= 10')
            )
        ],
    )
    def max_participants(self, max_participants):
        self.__max_participants = max_participants

    def set_max_participants(self, max_participants):
        self.max_participants = max_participants
        return self

    @property
    def wait_music_url(self):
        return self.__wait_music_url

    @wait_music_url.setter
    @validate_args(wait_music_url=[optional(of_type_exact(str), is_url())])
    def wait_music_url(self, wait_music_url):
        self.__wait_music_url = wait_music_url

    def set_wait_music_url(self, wait_music_url):
        self.wait_music_url = wait_music_url
        return self

    @property
    def wait_music_method(self):
        return self.__wait_music_method

    @wait_music_method.setter
    @validate_args(wait_music_method=[optional(of_type_exact(str),
                                               is_in(('GET', 'POST'), case_sensitive=False))])
    def wait_music_method(self, wait_music_method):
        self.__wait_music_method = wait_music_method

    def set_wait_music_method(self, wait_music_method):
        self.wait_music_method = wait_music_method
        return self

    @property
    def agent_hold_music_url(self):
        return self.__agent_hold_music_url

    @agent_hold_music_url.setter
    @validate_args(agent_hold_music_url=[optional(of_type_exact(str), is_url())])
    def agent_hold_music_url(self, agent_hold_music_url):
        self.__agent_hold_music_url = agent_hold_music_url

    def set_agent_hold_music_url(self, agent_hold_music_url):
        self.agent_hold_music_url = agent_hold_music_url
        return self

    @property
    def agent_hold_music_method(self):
        return self.__agent_hold_music_method

    @agent_hold_music_method.setter
    @validate_args(agent_hold_music_method=[optional(of_type_exact(str),
                                                     is_in(('GET', 'POST'), case_sensitive=False))])
    def agent_hold_music_method(self, agent_hold_music_method):
        self.__agent_hold_music_method = agent_hold_music_method

    def set_agent_hold_music_method(self, agent_hold_music_method):
        self.agent_hold_music_method = agent_hold_music_method
        return self

    @property
    def customer_hold_music_url(self):
        return self.__customer_hold_music_url

    @customer_hold_music_url.setter
    @validate_args(customer_hold_music_url=[optional(of_type_exact(str), is_url())])
    def customer_hold_music_url(self, customer_hold_music_url):
        self.__customer_hold_music_url = customer_hold_music_url

    def set_customer_hold_music_url(self, customer_hold_music_url):
        self.customer_hold_music_url = customer_hold_music_url
        return self

    @property
    def customer_hold_music_method(self):
        return self.__customer_hold_music_method

    @customer_hold_music_method.setter
    @validate_args(customer_hold_music_method=[optional(of_type_exact(str),
                                                        is_in(('GET', 'POST'), case_sensitive=False))])
    def customer_hold_music_method(self, customer_hold_music_method):
        self.__customer_hold_music_method = customer_hold_music_method

    def set_customer_hold_music_method(self, customer_hold_music_method):
        self.customer_hold_music_method = customer_hold_music_method
        return self

    @property
    def record(self):
        return self.__record

    @record.setter
    @validate_args(record=[optional(of_type_exact(bool))])
    def record(self, record):
        self.__record = record

    def set_record(self, record):
        self.record = record
        return self

    @property
    def record_file_format(self):
        return self.__record_file_format

    @record_file_format.setter
    @validate_args(
        record_file_format=[
            optional(
                of_type_exact(str),
                is_in(('mp3', 'wav'), case_sensitive=False, case_type='lower')
            )
        ]
    )
    def record_file_format(self, record_file_format):
        self.__record_file_format = record_file_format

    def set_record_file_format(self, record_file_format):
        self.record_file_format = record_file_format
        return self

    @property
    def recording_callback_url(self):
        return self.__recording_callback_url

    @recording_callback_url.setter
    @validate_args(recording_callback_url=[optional(of_type_exact(str), is_url())])
    def recording_callback_url(self, recording_callback_url):
        self.__recording_callback_url = recording_callback_url

    def set_recording_callback_url(self, recording_callback_url):
        self.recording_callback_url = recording_callback_url
        return self

    @property
    def recording_callback_method(self):
        return self.__recording_callback_method

    @recording_callback_method.setter
    @validate_args(recording_callback_method=[optional(of_type_exact(str),
                                                       is_in(('GET', 'POST'), case_sensitive=False))])
    def recording_callback_method(self, recording_callback_method):
        self.__recording_callback_method = recording_callback_method

    def set_recording_callback_method(self, recording_callback_method):
        self.recording_callback_method = recording_callback_method
        return self

    @property
    def status_callback_events(self):
        return self.__status_callback_events

    @status_callback_events.setter
    @validate_args(
        status_callback_events=[
            optional(
                of_type_exact(str),
                multi_is_in(('mpc-state-changes',
                             'participant-state-changes',
                             'participant-speak-events',
                             'participant-digit-input-events',
                             'add-participant-api-events'),
                            case_sensitive=False,
                            make_lower_case=True)
            )
        ]
    )
    def status_callback_events(self, status_callback_events):
        self.__status_callback_events = status_callback_events

    def set_status_callback_events(self, status_callback_events):
        self.status_callback_events = status_callback_events
        return self

    @property
    def status_callback_url(self):
        return self.__status_callback_url

    @status_callback_url.setter
    @validate_args(status_callback_url=[optional(of_type_exact(str), is_url())])
    def status_callback_url(self, status_callback_url):
        self.__status_callback_url = status_callback_url

    def set_status_callback_url(self, status_callback_url):
        self.status_callback_url = status_callback_url
        return self

    @property
    def status_callback_method(self):
        return self.__status_callback_method

    @status_callback_method.setter
    @validate_args(status_callback_method=[optional(of_type_exact(str),
                                                    is_in(('GET', 'POST'), case_sensitive=False))])
    def status_callback_method(self, status_callback_method):
        self.__status_callback_method = status_callback_method

    def set_status_callback_method(self, status_callback_method):
        self.status_callback_method = status_callback_method
        return self

    @property
    def stay_alone(self):
        return self.__stay_alone

    @stay_alone.setter
    @validate_args(stay_alone=[optional(of_type_exact(bool))])
    def stay_alone(self, stay_alone):
        self.__stay_alone = stay_alone

    def set_stay_alone(self, stay_alone):
        self.stay_alone = stay_alone
        return self

    @property
    def role(self):
        return self.__role

    @role.setter
    @validate_args(
        role=[
            of_type_exact(str),
            is_in(('agent', 'supervisor', 'customer'), case_sensitive=False, case_type='lower')
        ]
    )
    def role(self, role):
        self.__role = role

    def set_role(self, role):
        self.role = role
        return self

    @property
    def coach_mode(self):
        return self.__coach_mode

    @coach_mode.setter
    @validate_args(coach_mode=[optional(of_type_exact(bool))])
    def coach_mode(self, coach_mode):
        self.__coach_mode = coach_mode

    def set_coach_mode(self, coach_mode):
        self.coach_mode = coach_mode
        return self

    @property
    def mute(self):
        return self.__mute

    @mute.setter
    @validate_args(mute=[optional(of_type_exact(bool))])
    def mute(self, mute):
        self.__mute = mute

    def set_mute(self, mute):
        self.mute = mute
        return self

    @property
    def hold(self):
        return self.__hold

    @hold.setter
    @validate_args(hold=[optional(of_type_exact(bool))])
    def hold(self, hold):
        self.__hold = hold

    def set_hold(self, hold):
        self.hold = hold
        return self

    @property
    def start_mpc_on_enter(self):
        return self.__start_mpc_on_enter

    @start_mpc_on_enter.setter
    @validate_args(start_mpc_on_enter=[optional(of_type_exact(bool))])
    def start_mpc_on_enter(self, start_mpc_on_enter):
        self.__start_mpc_on_enter = start_mpc_on_enter

    def set_start_mpc_on_enter(self, start_mpc_on_enter):
        self.start_mpc_on_enter = start_mpc_on_enter
        return self

    @property
    def end_mpc_on_exit(self):
        return self.__end_mpc_on_exit

    @end_mpc_on_exit.setter
    @validate_args(end_mpc_on_exit=[optional(of_type_exact(bool))])
    def end_mpc_on_exit(self, end_mpc_on_exit):
        self.__end_mpc_on_exit = end_mpc_on_exit

    def set_end_mpc_on_exit(self, end_mpc_on_exit):
        self.end_mpc_on_exit = end_mpc_on_exit
        return self

    @property
    def enter_sound(self):
        return self.__enter_sound

    @enter_sound.setter
    @validate_args(
        enter_sound=[
            optional(
                all_of(
                    of_type_exact(str),
                    one_of(is_url(), is_in(('beep:1', 'beep:2', 'none'), case_sensitive=False, case_type='lower'))
                )
            )
        ]
    )
    def enter_sound(self, enter_sound):
        self.__enter_sound = enter_sound

    def set_enter_sound(self, enter_sound):
        self.enter_sound = enter_sound
        return self

    @property
    def enter_sound_method(self):
        return self.__enter_sound_method

    @enter_sound_method.setter
    @validate_args(enter_sound_method=[optional(of_type_exact(str),
                                                is_in(('GET', 'POST'), case_sensitive=False))])
    def enter_sound_method(self, enter_sound_method):
        self.__enter_sound_method = enter_sound_method

    def set_enter_sound_method(self, enter_sound_method):
        self.enter_sound_method = enter_sound_method
        return self

    @property
    def exit_sound(self):
        return self.__exit_sound

    @exit_sound.setter
    @validate_args(
        exit_sound=[
            optional(
                all_of(
                    of_type_exact(str),
                    one_of(is_url(), is_in(('beep:1', 'beep:2', 'none'), case_sensitive=False, case_type='lower'))
                )
            )
        ]
    )
    def exit_sound(self, exit_sound):
        self.__exit_sound = exit_sound

    def set_exit_sound(self, exit_sound):
        self.exit_sound = exit_sound
        return self

    @property
    def exit_sound_method(self):
        return self.__exit_sound_method

    @exit_sound_method.setter
    @validate_args(exit_sound_method=[optional(of_type_exact(str),
                                               is_in(('GET', 'POST'), case_sensitive=False))])
    def exit_sound_method(self, exit_sound_method):
        self.__exit_sound_method = exit_sound_method

    def set_exit_sound_method(self, exit_sound_method):
        self.exit_sound_method = exit_sound_method
        return self

    @property
    def on_exit_action_url(self):
        return self.__on_exit_action_url

    @on_exit_action_url.setter
    @validate_args(on_exit_action_url=[optional(of_type_exact(str), is_url())])
    def on_exit_action_url(self, on_exit_action_url):
        self.__on_exit_action_url = on_exit_action_url

    def set_on_exit_action_url(self, on_exit_action_url):
        self.on_exit_action_url = on_exit_action_url
        return self

    @property
    def on_exit_action_method(self):
        return self.__on_exit_action_method

    @on_exit_action_method.setter
    @validate_args(on_exit_action_method=[optional(of_type_exact(str),
                                                   is_in(('GET', 'POST'), case_sensitive=False))])
    def on_exit_action_method(self, on_exit_action_method):
        self.__on_exit_action_method = on_exit_action_method

    def set_on_exit_action_method(self, on_exit_action_method):
        self.on_exit_action_method = on_exit_action_method
        return self

    @property
    def relay_dtmf_inputs(self):
        return self.__relay_dtmf_inputs

    @relay_dtmf_inputs.setter
    @validate_args(relay_dtmf_inputs=[optional(of_type_exact(bool))])
    def relay_dtmf_inputs(self, relay_dtmf_inputs):
        self.__relay_dtmf_inputs = relay_dtmf_inputs

    def set_relay_dtmf_inputs(self, relay_dtmf_inputs):
        self.relay_dtmf_inputs = relay_dtmf_inputs
        return self

    def __init__(
            self,
            content,
            role,
            max_duration=14400,
            max_participants=10,
            wait_music_url=None,
            wait_music_method='GET',
            agent_hold_music_url=None,
            agent_hold_music_method='GET',
            customer_hold_music_url=None,
            customer_hold_music_method='GET',
            record=False,
            record_file_format='mp3',
            recording_callback_url=None,
            recording_callback_method='POST',
            status_callback_events='mpc-state-changes,participant-state-changes',
            status_callback_url=None,
            status_callback_method='POST',
            stay_alone=False,
            coach_mode=True,
            mute=False,
            hold=False,
            start_mpc_on_enter=True,
            end_mpc_on_exit=False,
            enter_sound='beep:1',
            enter_sound_method='GET',
            exit_sound='beep:2',
            exit_sound_method='GET',
            on_exit_action_url=None,
            on_exit_action_method='POST',
            relay_dtmf_inputs=False,
    ):
        super(MultiPartyCallElement, self).__init__()
        self.content = content
        self.role = role
        self.max_duration = max_duration
        self.max_participants = max_participants
        self.wait_music_url = wait_music_url
        self.wait_music_method = wait_music_method
        self.agent_hold_music_url = agent_hold_music_url
        self.agent_hold_music_method = agent_hold_music_method
        self.customer_hold_music_url = customer_hold_music_url
        self.customer_hold_music_method = customer_hold_music_method
        self.record = record
        self.record_file_format = record_file_format
        self.recording_callback_url = recording_callback_url
        self.recording_callback_method = recording_callback_method
        self.status_callback_events = status_callback_events
        self.status_callback_url = status_callback_url
        self.status_callback_method = status_callback_method
        self.stay_alone = stay_alone
        self.coach_mode = coach_mode
        self.mute = mute
        self.hold = hold
        self.start_mpc_on_enter = start_mpc_on_enter
        self.end_mpc_on_exit = end_mpc_on_exit
        self.enter_sound = enter_sound
        self.enter_sound_method = enter_sound_method
        self.exit_sound = exit_sound
        self.exit_sound_method = exit_sound_method
        self.on_exit_action_url = on_exit_action_url
        self.on_exit_action_method = on_exit_action_method
        self.relay_dtmf_inputs = relay_dtmf_inputs

    def to_dict(self):
        d = {
            'role': self.role,
            'maxDuration': self.max_duration,
            'maxParticipants': self.max_participants,
            'waitMusicUrl': self.wait_music_url,
            'waitMusicMethod': self.wait_music_method,
            'agentHoldMusicUrl': self.agent_hold_music_url,
            'agentHoldMusicMethod': self.agent_hold_music_method,
            'customerHoldMusicUrl': self.customer_hold_music_url,
            'customerHoldMusicMethod': self.customer_hold_music_method,
            'record': self.record,
            'recordFileFormat': self.record_file_format,
            'recordingCallbackUrl': self.recording_callback_url,
            'recordingCallbackMethod': self.recording_callback_method,
            'statusCallbackEvents': self.status_callback_events,
            'statusCallbackUrl': self.status_callback_url,
            'statusCallbackMethod': self.status_callback_method,
            'stayAlone': self.stay_alone,
            'coachMode': self.coach_mode,
            'mute': self.mute,
            'hold': self.hold,
            'startMpcOnEnter': self.start_mpc_on_enter,
            'endMpcOnExit': self.end_mpc_on_exit,
            'enterSound': self.enter_sound,
            'enterSoundMethod': self.enter_sound_method,
            'exitSound': self.exit_sound,
            'exitSoundMethod': self.exit_sound_method,
            'onExitActionUrl': self.on_exit_action_url,
            'onExitActionMethod': self.on_exit_action_method,
            'relayDTMFInputs': self.relay_dtmf_inputs
        }
        return {
            k: six.text_type(map_type(v))
            for k, v in d.items() if v is not None
        }



# -*- coding: utf-8 -*-
"""
Core client, used for all API requests.
"""

import os
import platform
from collections import namedtuple

from plivo.base import ResponseObject
from plivo.exceptions import (AuthenticationError, InvalidRequestError,
                              PlivoRestError, PlivoServerError,
                              ResourceNotFoundError, ValidationError)
from plivo.resources import (Accounts, Addresses, Applications, Calls,
                             Conferences, Endpoints, Identities,
                             Messages, Powerpacks, Media, Lookup,
                             Numbers, Pricings, Recordings, Subaccounts, CallFeedback, MultiPartyCalls)
from plivo.resources.live_calls import LiveCalls
from plivo.resources.queued_calls import QueuedCalls
from plivo.resources.regulatory_compliance import EndUsers, ComplianceDocumentTypes, ComplianceDocuments, \
    ComplianceRequirements, ComplianceApplications
from plivo.utils import is_valid_mainaccount, is_valid_subaccount
from plivo.version import __version__
from requests import Request, Session

AuthenticationCredentials = namedtuple('AuthenticationCredentials',
                                       'auth_id auth_token')

PLIVO_API = 'https://api.plivo.com'
PLIVO_API_BASE_URI = '/'.join([PLIVO_API, 'v1/Account'])

# Will change these urls before putting this change in production
API_VOICE = 'https://voice.plivo.com'
API_VOICE_BASE_URI = '/'.join([API_VOICE, 'v1/Account'])
API_VOICE_FALLBACK_1 = 'https://voice-usw1.plivo.com'
API_VOICE_FALLBACK_2 = 'https://voice-use1.plivo.com'
API_VOICE_BASE_URI_FALLBACK_1 = '/'.join([API_VOICE_FALLBACK_1, 'v1/Account'])
API_VOICE_BASE_URI_FALLBACK_2 = '/'.join([API_VOICE_FALLBACK_2, 'v1/Account'])

CALLINSIGHTS_BASE_URL = 'https://stats.plivo.com'


def get_user_agent():
    return 'plivo-python/%s (Python: %s)' % (__version__,
                                             platform.python_version())


def fetch_credentials(auth_id, auth_token):
    """Fetches the right credentials either from params or from environment"""

    if not (auth_id and auth_token):
        try:
            auth_id = os.environ['PLIVO_AUTH_ID']
            auth_token = os.environ['PLIVO_AUTH_TOKEN']
        except KeyError:
            raise AuthenticationError('The Plivo Python SDK '
                                      'could not find your auth credentials.')

    if not (is_valid_mainaccount(auth_id) or is_valid_subaccount(auth_id)):
        raise AuthenticationError('Invalid auth_id supplied: %s' % auth_id)

    return AuthenticationCredentials(auth_id=auth_id, auth_token=auth_token)


class Client(object):
    def __init__(self, auth_id=None, auth_token=None, proxies=None, timeout=5):
        """
        The Plivo API client.

        Deals with all the API requests to be made.
        """

        self.base_uri = PLIVO_API_BASE_URI
        self.session = Session()
        self.session.headers.update({
            'User-Agent': get_user_agent(),
            'Content-Type': 'application/json',
            'Accept': 'application/json',
        })
        self.session.auth = fetch_credentials(auth_id, auth_token)
        self.multipart_session = Session()
        self.multipart_session.headers.update({
            'User-Agent': get_user_agent(),
            'Cache-Control': 'no-cache',
        })
        self.multipart_session.auth = fetch_credentials(auth_id, auth_token)
        self.proxies = proxies
        self.timeout = timeout
        self.account = Accounts(self)
        self.subaccounts = Subaccounts(self)
        self.applications = Applications(self)
        self.calls = Calls(self)
        self.live_calls = LiveCalls(self)
        self.queued_calls = QueuedCalls(self)
        self.conferences = Conferences(self)
        self.endpoints = Endpoints(self)
        self.messages = Messages(self)
        self.lookup = Lookup(self)
        self.numbers = Numbers(self)
        self.powerpacks = Powerpacks(self)
        self.media = Media(self)
        self.pricing = Pricings(self)
        self.recordings = Recordings(self)
        self.addresses = Addresses(self)
        self.identities = Identities(self)
        self.call_feedback = CallFeedback(self)
        self.end_users = EndUsers(self)
        self.compliance_document_types = ComplianceDocumentTypes(self)
        self.compliance_documents = ComplianceDocuments(self)
        self.compliance_requirements = ComplianceRequirements(self)
        self.compliance_applications = ComplianceApplications(self)
        self.multi_party_calls = MultiPartyCalls(self)
        self.voice_retry_count = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()
        self.multipart_session.close()

    def process_response(self,
                         method,
                         response,
                         response_type=None,
                         objects_type=None):
        """Processes the API response based on the status codes and method used
        to access the API
        """
        try:
            response_json = response.json(
                object_hook=lambda x: ResponseObject(x) if isinstance(x, dict) else x)
            if response_type:
                r = response_type(self, response_json.__dict__)
                response_json = r

            if 'objects' in response_json and objects_type:
                response_json.objects = [
                    objects_type(self, obj.__dict__)
                    for obj in response_json.objects
                ]
        except ValueError:
            response_json = None

        if response.status_code == 400:
            if response_json and 'error' in response_json:
                raise ValidationError(response_json.error)
            raise ValidationError(
                'A parameter is missing or is invalid while accessing resource'
                'at: {url}'.format(url=response.url))

        if response.status_code == 401:
            if response_json and 'error' in response_json:
                raise AuthenticationError(response_json.error)
            raise AuthenticationError(
                'Failed to authenticate while accessing resource at: '
                '{url}'.format(url=response.url))

        if response.status_code == 404:
            if response_json and 'error' in response_json:
                raise ResourceNotFoundError(response_json.error)
            raise ResourceNotFoundError(
                'Resource not found at: {url}'.format(url=response.url))

        if response.status_code == 405:
            if response_json and 'error' in response_json:
                raise InvalidRequestError(response_json.error)
            raise InvalidRequestError(
                'HTTP method "{method}" not allowed to access resource at: '
                '{url}'.format(method=method, url=response.url))

        if response.status_code == 409:
            if response_json and 'error' in response_json:
                raise InvalidRequestError(response_json.error)
            raise InvalidRequestError(
                'Conflict: '
                '{url}'.format(url=response.url))

        if response.status_code == 422:
            if response_json and 'error' in response_json:
                raise InvalidRequestError(response_json.error)
            raise InvalidRequestError(
                'Unprocessable Entity: '
                '{url}'.format(url=response.url))

        if response.status_code == 500:
            if response_json and 'error' in response_json:
                raise PlivoServerError(response_json.error)
            raise PlivoServerError(
                'A server error occurred while accessing resource at: '
                '{url}'.format(url=response.url))

        if method == 'DELETE':
            if response.status_code not in [200, 204]:
                raise PlivoRestError('Resource at {url} could not be '
                                     'deleted'.format(url=response.url))

        elif response.status_code not in [200, 201, 202, 204, 207]:
            raise PlivoRestError(
                'Received status code {status_code} for the HTTP method '
                '"{method}"'.format(
                    status_code=response.status_code, method=method))

        self.voice_retry_count = 0

        return response_json

    def create_request(self, method, path=None, data=None, **kwargs):
        # The abstraction created by request() and create_request() is moot
        # now since several product-specific handling have been aded.
        # Requires a refactor.
        if 'is_callinsights_request' in kwargs:
            url = '/'.join([CALLINSIGHTS_BASE_URL, kwargs['callinsights_request_path']])
            req = Request(method, url, **({'params': data} if method == 'GET' else {'json': data}))
        elif kwargs.get('is_lookup_request', False):
            path = path or []
            url = '/'.join(list([str(p) for p in path]))
            req = Request(method, url, **({'params': data} if method == 'GET' else {'json': data}))
        else:
            path = path or []
            req = Request(method, '/'.join([self.base_uri, self.session.auth[0]] +
                                           list([str(p) for p in path])) + '/',
                          **({
                                 'params': data
                             } if method == 'GET' else {
                              'json': data
                          }))
        return self.session.prepare_request(req)

    def create_multipart_request(self,
                                 method,
                                 path=None,
                                 data=None,
                                 files=None):
        path = path or []

        data_args = {}
        if method == 'GET':
            data_args['params'] = data
        else:
            data_args['data'] = data
            try:
                if files:
                    data_args['files'] = files
            except Exception as e:
                print(e)
        url = '/'.join([self.base_uri, self.multipart_session.auth[0]] + list([str(p) for p in path])) + '/'
        req = Request(method, url, **data_args)
        return self.multipart_session.prepare_request(req)

    def send_request(self, request, **kwargs):
        if 'session' in kwargs:
            session = kwargs['session']
            del kwargs['session']
        else:
            session = self.session
        return session.send(
            request, proxies=self.proxies, timeout=self.timeout, **kwargs)

    def request(self,
                method,
                path=None,
                data=None,
                response_type=None,
                objects_type=None,
                files=None,
                **kwargs):
        if files is not None:
            req = self.create_multipart_request(method, path, data, files)
            session = self.multipart_session
        else:
            if not kwargs.get("is_voice_request", False):
                self.base_uri = PLIVO_API_BASE_URI
            if data and 'is_callinsights_request' in data:
                params_dict = {}
                if 'callinsights_request_path' in data:
                    params_dict['is_callinsights_request'] = data['is_callinsights_request']
                    params_dict['callinsights_request_path'] = data['callinsights_request_path']
                    del data['is_callinsights_request']
                    del data['callinsights_request_path']
                    req = self.create_request(method, path, data, **params_dict)
            elif kwargs.get("is_voice_request", False):
                del kwargs["is_voice_request"]
                if self.voice_retry_count == 0:
                    self.base_uri = API_VOICE_BASE_URI
                req = self.create_request(method, path, data)
                session = self.session
                kwargs['session'] = session
                response = self.send_request(req, **kwargs)
                if response.status_code >= 500:
                    print('Fallback for URL: {}. Retry {}'.format(response.url, self.voice_retry_count))
                    self.voice_retry_count += 1
                    if self.voice_retry_count == 1:
                        self.base_uri = API_VOICE_BASE_URI_FALLBACK_1
                    elif self.voice_retry_count == 2:
                        self.base_uri = API_VOICE_BASE_URI_FALLBACK_2
                    else:
                        return self.process_response(method, response, response_type, objects_type)
                    kwargs["is_voice_request"] = True
                    return self.request(method, path, data, **kwargs)
                return self.process_response(method, response, response_type, objects_type)
            elif kwargs.get('is_lookup_request', False):
                req = self.create_request(method, path, data, is_lookup_request=True)
                del kwargs['is_lookup_request']
            else:
                req = self.create_request(method, path, data)
            session = self.session
        kwargs['session'] = session
        res = self.send_request(req, **kwargs)
        return self.process_response(method, res, response_type, objects_type)



# -*- coding: utf-8 -*-
from plivo.base import PlivoResource,\
    PlivoResourceInterface
from plivo.utils import to_param_dict
from ..utils.validators import *
from plivo.utils import to_param_dict


class LiveCall(PlivoResource):
    _name = 'LiveCall'
    _identifier_string = 'call_uuid'


class LiveCalls(PlivoResourceInterface):
    _resource_type = LiveCall
    _iterable = False

    @validate_args(
        limit=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(
                        lambda limit: 0 < limit <= 20,
                        message='0 < limit <= 20')))
        ],
        offset=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda offset: 0 <= offset, message='0 <= offset')))
        ],
        call_direction=[
            optional(of_type(six.text_type), is_in(('inbound', 'outbound')))
        ],
        from_number=[optional(is_phonenumber())],
        to_number=[optional(is_iterable(of_type(six.text_type), sep='<'))]
    )
    def list_ids(self,
                 call_direction=None,
                 from_number=None,
                 to_number=None,
                 limit=20,
                 offset=0,
                 ):
        params = to_param_dict(self.list_ids, locals())
        params.update({'status': 'live'})
        return self.client.request('GET', ('Call',), params, is_voice_request=True)

    @validate_args(_id=[of_type(six.text_type)])
    def get(self, _id):
        return self.client.request(
            'GET', ('Call', _id), {'status': 'live'}, is_voice_request=True
        )



from lxml import etree

from plivo.exceptions import PlivoXMLError


class PlivoXMLElement(object):
    def __init__(self):
        self.content = ''
        self.children = []

    def add(self, element):
        if not isinstance(element, PlivoXMLElement):
            raise PlivoXMLError('element must be a PlivoXMLElement')

        if element._name not in self._nestable:
            raise PlivoXMLError(
                '{} is not nestable in {} (allowed: {})'.format(
                    element._name, self._name, self._nestable))
        self.children.append(element)
        return self

    def continue_speak(self, body=None):
        return body.replace('<cont>', ' ').replace('</cont>', ' ')

    def to_string(self, pretty=True):
        s = self.continue_speak(etree.tostring(self._to_element(), pretty_print=pretty, encoding='unicode'))

        if not isinstance(s, str):
            s = s.encode('utf-8')
        return s

    def _to_element(self, parent=None):
        e = etree.SubElement(
            parent, self._name,
            **self.to_dict()) if parent is not None else etree.Element(
                self._name, **self.to_dict())
        if self.content:
            try:
                e.text = self.content.decode()
            except:
                e.text = self.content
        for child in self.children:
            child._to_element(parent=e)
        return e



# -*- coding: utf-8 -*-
"""
Application class - along with its list class
"""

from plivo.base import (ListResponseObject, PlivoResource,
                        PlivoResourceInterface)
from plivo.resources.accounts import Subaccount
from plivo.utils import to_param_dict
from plivo.utils.validators import *


class Application(PlivoResource):
    _name = 'Application'
    _identifier_string = 'app_id'

    def update(self,
               answer_url,
               answer_method='POST',
               hangup_url=None,
               hangup_method='POST',
               fallback_answer_url=None,
               fallback_method='POST',
               message_url=None,
               message_method='POST',
               default_number_app=False,
               default_endpoint_app=False,
               subaccount=None,
			   log_incoming_messages=True,
               public_uri=None):
        params = to_param_dict(self.update, locals())
        self.__dict__.update(params)
        return self.client.applications.update(self.id, **params)

    def delete(self, cascade=None, new_endpoint_application=None):
        return self.client.applications.delete(self.id, cascade, new_endpoint_application)

    def get(self):
        resp = self.client.applications.get()
        self.__dict__.update(resp.__dict__)
        return resp


class Applications(PlivoResourceInterface):
    _resource_type = Application

    @validate_args(
        answer_url=[is_url()],
        app_name=[of_type(six.text_type)],
        answer_method=[optional(of_type(six.text_type))],
        hangup_url=[optional(is_url())],
        hangup_method=[optional(of_type(six.text_type))],
        fallback_answer_url=[optional(is_url())],
        fallback_method=[optional(of_type(six.text_type))],
        message_url=[optional(is_url())],
        message_method=[optional(of_type(six.text_type))],
        default_number_app=[optional(of_type_exact(bool))],
        default_endpoint_app=[optional(of_type_exact(bool))],
        subaccount=[optional(is_subaccount())],
        log_incoming_messages=[optional(of_type_exact(bool))],
        public_uri=[optional(of_type_exact(bool))])
    def create(self,
               answer_url,
               app_name,
               answer_method='POST',
               hangup_url=None,
               hangup_method='POST',
               fallback_answer_url=None,
               fallback_method='POST',
               message_url=None,
               message_method='POST',
               default_number_app=False,
               default_endpoint_app=False,
               subaccount=None,
               log_incoming_messages=True,
               public_uri=None):

        if subaccount:
            if isinstance(subaccount, Subaccount):
                subaccount = subaccount.id
        return self.client.request('POST', ('Application', ), to_param_dict(self.create, locals()), is_voice_request=True)

    @validate_args(app_id=[of_type(six.text_type)])
    def get(self, app_id):
        return self.client.request(
            'GET', ('Application', app_id), response_type=Application, is_voice_request=True)

    @validate_args(
        subaccount=[optional(is_subaccount())],
        limit=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda limit: 0 < limit <= 20, '0 < limit <= 20')))
        ],
        offset=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda offset: 0 <= offset, '0 <= offset')))
        ])
    def list(self, subaccount=None, limit=20, offset=0):
        if subaccount:
            if isinstance(subaccount, Subaccount):
                subaccount = subaccount.id
        return self.client.request(
            'GET', ('Application', ),
            to_param_dict(self.list, locals()),
            response_type=ListResponseObject,
            objects_type=Application, is_voice_request=True)

    @validate_args(
        answer_url=[is_url()],
        app_id=[of_type(six.text_type)],
        answer_method=[optional(of_type(six.text_type))],
        hangup_url=[optional(is_url())],
        hangup_method=[optional(of_type(six.text_type))],
        fallback_answer_url=[optional(is_url())],
        fallback_method=[optional(of_type(six.text_type))],
        message_url=[optional(is_url())],
        message_method=[optional(of_type(six.text_type))],
        default_number_app=[optional(of_type_exact(bool))],
        default_endpoint_app=[optional(of_type_exact(bool))],
        subaccount=[optional(is_subaccount())],
        log_incoming_messages=[optional(of_type_exact(bool))],
        public_uri=[optional(of_type_exact(bool))])
    def update(self,
               app_id,
               answer_url,
               answer_method='POST',
               hangup_url=None,
               hangup_method='POST',
               fallback_answer_url=None,
               fallback_method='POST',
               message_url=None,
               message_method='POST',
               default_number_app=False,
               default_endpoint_app=False,
               subaccount=None,
               log_incoming_messages=True,
               public_uri=None):
        if subaccount:
            if isinstance(subaccount, Subaccount):
                subaccount = subaccount.id
        return self.client.request('POST', ('Application', app_id),
                                   to_param_dict(self.update, locals()), is_voice_request=True)

    @validate_args(
        app_id=[of_type(six.text_type)],
        new_endpoint_application=[optional(of_type(six.text_type))],
        cascade=[optional(of_type_exact(bool))]
    )
    def delete(self, app_id, cascade=None, new_endpoint_application=None):
        return self.client.request('DELETE', ('Application', app_id),
                                   to_param_dict(self.delete, locals()), is_voice_request=True)



# -*- coding: utf-8 -*-
from plivo.utils.validators import *

from ..base import ListResponseObject, PlivoResource, PlivoResourceInterface
from ..exceptions import *
from ..utils import *


class Message(PlivoResource):
    _name = 'Message'
    _identifier_string = 'message_uuid'

    def get(self):
        return self.client.messages.get(self.id)

    def delete(self):
        raise InvalidRequestError('Cannot delete a Message resource')

    def update(self):
        raise InvalidRequestError('Cannot update a Message resource')

    def listMedia(self):
        return self.client.request(
            'GET', ('Message', self.id, 'Media'), response_type=None)


class Messages(PlivoResourceInterface):
    _resource_type = Message

    @validate_args(
        src=[optional(is_phonenumber())],
        dst=[is_iterable(of_type(six.text_type), '<')],
        text=[optional(of_type(six.text_type))],
        type_=[
            optional(all_of(of_type(six.text_type), is_in(('sms', 'mms'))))],
        url=[optional(is_url())],
        method=[optional(of_type(six.text_type))],
        log=[optional(of_type_exact(bool))],
        trackable=[optional(of_type_exact(bool))],
        powerpack_uuid=[optional(of_type(six.text_type))],
        media_urls=[optional(of_type_exact(list))],
        media_ids=[optional(of_type_exact(list))])
    def create(self,
               dst,
               text=None,
               src=None,
               type_='sms',
               url=None,
               method='POST',
               log=True,
               trackable=False,
               powerpack_uuid=None,
               media_urls=None,
               media_ids=None):
        if src in dst.split('<'):
            raise ValidationError(
                'destination number cannot be same as source number')
        elif (src is None) and (powerpack_uuid is None):
            raise ValidationError(
                'Specify either powerpack_uuid or src in request params to send a message'
            )
        elif (src is not None) and (powerpack_uuid is not None):
            raise ValidationError(
                'Both powerpack_uuid and src cannot be specified. Specify either powerpack_uuid or src in request params to send a message.'
            )
        return self.client.request('POST', ('Message', ),
                                   to_param_dict(self.create, locals()))

    @validate_args(message_uuid=[of_type(six.text_type)])
    def get(self, message_uuid):
        return self.client.request(
            'GET', ('Message', message_uuid), response_type=Message)

    @validate_args(
        subaccount=[optional(is_subaccount())],
        message_direction=[optional(is_in(('inbound', 'outbound')))],
        message_time__gt=[optional(is_valid_date())],
        message_time__gte=[optional(is_valid_date())],
        message_time__lt=[optional(is_valid_date())],
        message_time__lte=[optional(is_valid_date())],
        message_time=[optional(is_valid_date())],
        message_state=[
            optional(
                is_in(('queued', 'sent', 'failed', 'delivered', 'undelivered',
                       'rejected')))
        ],
        limit=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda limit: 0 < limit <= 20, '0 < limit <= 20')))
        ],
        offset=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda offset: 0 <= offset, '0 <= offset')))
        ])
    def list(self,
             subaccount=None,
             message_direction=None,
             message_time__gt=None,
             message_time__gte=None,
             message_time__lt=None,
             message_time__lte=None,
             message_time=None,
             message_state=None,
             limit=None,
             offset=None,
             error_code=None):
        return self.client.request(
            'GET', ('Message', ),
            to_param_dict(self.list, locals()),
            response_type=ListResponseObject,
            objects_type=Message)



# -*- coding: utf-8 -*-
from plivo.utils.validators import *

from ..base import ListResponseObject, PlivoResource, PlivoResourceInterface
from ..exceptions import *
from ..utils import *


class Media(PlivoResource):
    _name = 'Media'
    _identifier_string = 'media_id'

    def get(self):
        return self.client.medias.get(self.id)


class Media(PlivoResourceInterface):
    _resource_type = Media

    @validate_args(
        media_file=[optional(of_type_exact(list))])
    def upload(self, media_file):
        if media_file:
            fileList = []
            for media_url in media_file:
                file_extension = media_url.strip().split('.')[-1].lower()
                if file_extension not in ['jpeg', 'jpg', 'png', 'xcf', 'plain', 'pdf', 'mpeg', 'mp4']:
                    raise ValidationError(
                        'File format of the file to be uploaded should be one of JPG, JPEG, PNG or PDF'
                    )
                content_types = {
                    'jpeg': 'image/jpeg',
                    'jpg': 'image/jpeg',
                    'png': 'image/png',
                    'pdf': 'application/pdf',
                    'xcf': 'image/xcf',
                    'text': 'text/plain',
                    'mpeg': 'video/mpeg',
                    'mp4': 'video/mp4'
                }
                print(media_url)
                import os
                files = (
                    'file', (media_url.split(os.sep)[-1], open(
                        media_url, 'rb'), content_types[file_extension])
                )
                fileList.append(files)
        data_to_send = {}
        return self.client.request(
            'POST', ('Media', ), data_to_send, files=fileList)

    @validate_args(media_id=[of_type(six.text_type)])
    def get(self, media_id):
        return self.client.request(
            'GET', ('Media', media_id), response_type=None)

    @validate_args(
        limit=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda limit: 0 < limit <= 20, '0 < limit <= 20')))
        ],
        offset=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda offset: 0 <= offset, '0 <= offset')))
        ])
    def list(self,
             limit=20,
             offset=0):
        return self.client.request(
            'GET', ('Media', ),
            to_param_dict(self.list, locals()),
            response_type=None)



import jwt, time
from plivo.utils.validators import *

"""
Class to represent plivo token for endpoint authentication
"""
class AccessToken:
    auth_id = ''
    username = ''
    valid_from = 0
    lifetime = 86400
    key = ''
    grants = {}
    uid = 0

    @validate_args(
        auth_id=[is_account_id()],
        auth_token=[optional(of_type(six.text_type))],
        username=[all_of(
            of_type(six.text_type),
            check(lambda username: len(username) > 0, 'empty username')
        )],
        valid_from=[optional(of_type(*six.integer_types))],
        lifetime=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda lifetime: 180 <= lifetime <= 86400,
                          '180 < lifetime <= 86400')))
        ],
        valid_till=[optional(of_type(*six.integer_types))],
    )
    def __init__(self,
                 auth_id,
                 auth_token,
                 username,
                 valid_from=None,
                 lifetime=None,
                 valid_till=None,
                 uid=None):
        self.auth_id = auth_id
        self.username = username
        if valid_from:
            self.valid_from = int(valid_from)
        else:
            self.valid_from = int(time.time())
        if lifetime:
            self.lifetime = int(lifetime)
            if valid_till is not None:
                raise ValidationError("use either lifetime or valid_till")
        elif valid_till:
            self.lifetime = valid_till - self.valid_from
            if self.lifetime < 0:
                raise ValidationError(
                    "validity expires %s seconds before it starts" %
                    self.lifetime)
            if self.lifetime < 180 or self.lifetime > 86400:
                raise ValidationError(
                    "validity of %s seconds is out of permitted range [180, 86400]" %
                    self.lifetime)

        self.key = auth_token

        if uid:
            self.uid = uid
        else:
            self.uid = "%s-%s" % (username, time.time())

    @validate_args(
        incoming=[optional(of_type_exact(bool))],
        outgoing=[optional(of_type_exact(bool))],
    )
    def add_voice_grants(self, incoming=False, outgoing=False):
        self.grants['voice'] = {
            'incoming_allow': incoming,
            'outgoing_allow': outgoing
        }

    def to_jwt(self):
        headers = {'typ': 'JWT', 'cty': 'plivo;v=1'}
        algorithm = 'HS256'
        claims = {
            'jti': self.uid,
            'iss': self.auth_id,
            'sub': self.username,
            'nbf': self.valid_from,
            'exp': self.valid_from + self.lifetime,
            'grants': self.grants
        }
        return jwt.encode(claims, self.key, algorithm, headers).decode('utf-8')



# -*- coding: utf-8 -*-
"""
Node class
"""
from plivo.base import PlivoResource
from plivo.utils import to_param_dict
from plivo.utils.validators import *


class Node(PlivoResource):
    _name = 'Node'
    _identifier_string = 'node_id'

    @validate_args(
        action=[of_type(six.text_type)])
    def update(self,
               action,
               trigger_source,
               to,
               role):
        return self.client.request('POST', ('phlo', self.phlo_id, self.node_type, self.node_id),
                                   to_param_dict(self.update, locals()))


class MultiPartyCall(Node):
    _name = 'MultiPartyCall'
    _identifier_string = 'node_id'

    def call(self,
             trigger_source,
             to,
             role):
        return self.update('call', trigger_source, to, role)

    def warm_transfer(self,
             trigger_source,
             to,
             role='agent'):
        return self.update('warm_transfer', trigger_source, to, role)

    def cold_transfer(self,
             trigger_source,
             to,
             role='agent'):
        return self.update('cold_transfer', trigger_source, to, role)

    def member(self, member_id):
        self.member_id = member_id
        data = {
            'member_id': member_id,
            'phlo_id': self.phlo_id,
            'node_id': self.node_id,
            'node_type': self.node_type
        }
        return Member(self.client, data)


class Member(PlivoResource):
    _name = 'Member'
    _identifier_string = 'member_id'

    def abort_transfer(self):
        return self.update('abort_transfer')

    def resume_call(self):
        return self.update('resume_call')

    def voicemail_drop(self):
        return self.update('voicemail_drop')

    def hangup(self):
        return self.update('hangup')

    def hold(self):
        return self.update('hold')

    def unhold(self):
        return self.update('unhold')

    def mute(self):
        return self.update('mute')

    def unmute(self):
        return self.update('unmute')

    def update(self,
               action):
        return self.client.request('POST', ('phlo', self.phlo_id, self.node_type, self.node_id, 'members', self.member_id),
                                   to_param_dict(self.update, locals()))



from plivo.xml import (ConferenceElement, DialElement, DTMFElement,
                       GetDigitsElement, GetInputElement, HangupElement, 
                       MessageElement, PlayElement, PlivoXMLElement, 
                       PreAnswerElement, RecordElement, RedirectElement, 
                       SpeakElement, WaitElement, MultiPartyCallElement
                       )


class ResponseElement(PlivoXMLElement):
    _name = 'Response'
    _nestable = [
        'Conference',
        'Dial',
        'DTMF',
        'GetDigits',
        'GetInput',
        'Hangup',
        'Message',
        'Play',
        'PreAnswer',
        'Record',
        'Redirect',
        'Speak',
        'Wait',
        'MultiPartyCall',
    ]

    def add_conference(
            self,
            content,
            muted=None,
            enter_sound=None,
            exit_sound=None,
            start_conference_on_enter=None,
            end_conference_on_exit=None,
            stay_alone=None,
            wait_sound=None,
            max_members=None,
            record=None,
            record_file_format=None,
            time_limit=None,
            hangup_on_star=None,
            action=None,
            method=None,
            callback_url=None,
            callback_method=None,
            digits_match=None,
            floor_event=None,
            redirect=None,
            relay_dtmf=None,
    ):
        self.add(
            ConferenceElement(
                content=content,
                muted=muted,
                enter_sound=enter_sound,
                exit_sound=exit_sound,
                start_conference_on_enter=start_conference_on_enter,
                end_conference_on_exit=end_conference_on_exit,
                stay_alone=stay_alone,
                wait_sound=wait_sound,
                max_members=max_members,
                record=record,
                record_file_format=record_file_format,
                time_limit=time_limit,
                hangup_on_star=hangup_on_star,
                action=action,
                method=method,
                callback_url=callback_url,
                callback_method=callback_method,
                digits_match=digits_match,
                floor_event=floor_event,
                redirect=redirect,
                relay_dtmf=relay_dtmf,
            ))
        return self

    def add_dial(
            self,
            action=None,
            method=None,
            hangup_on_star=None,
            time_limit=None,
            timeout=None,
            caller_id=None,
            caller_name=None,
            confirm_sound=None,
            confirm_key=None,
            dial_music=None,
            callback_url=None,
            callback_method=None,
            redirect=None,
            digits_match=None,
            digits_match_b_leg=None,
            sip_headers=None,
    ):
        self.add(
            DialElement(
                action=action,
                method=method,
                hangup_on_star=hangup_on_star,
                time_limit=time_limit,
                timeout=timeout,
                caller_id=caller_id,
                caller_name=caller_name,
                confirm_sound=confirm_sound,
                confirm_key=confirm_key,
                dial_music=dial_music,
                callback_url=callback_url,
                callback_method=callback_method,
                redirect=redirect,
                digits_match=digits_match,
                digits_match_b_leg=digits_match_b_leg,
                sip_headers=sip_headers,
            ))
        return self

    def add_dtmf(
            self,
            content,
            async_=None,
    ):
        self.add(DTMFElement(
            content=content,
            async_=async_,
        ))
        return self

    def add_get_digits(
            self,
            action=None,
            method=None,
            timeout=None,
            digit_timeout=None,
            finish_on_key=None,
            num_digits=None,
            retries=None,
            redirect=None,
            play_beep=None,
            valid_digits=None,
            invalid_digits_sound=None,
            log=None,
    ):
        self.add(
            GetDigitsElement(
                action=action,
                method=method,
                timeout=timeout,
                digit_timeout=digit_timeout,
                finish_on_key=finish_on_key,
                num_digits=num_digits,
                retries=retries,
                redirect=redirect,
                play_beep=play_beep,
                valid_digits=valid_digits,
                invalid_digits_sound=invalid_digits_sound,
                log=log,
            ))
        return self

    def add_get_input(
            self,
            action=None,
            method=None,
            input_type=None,
            execution_timeout=None,
            digit_end_timeout=None,
            speech_end_timeout=None,
            finish_on_key=None,
            num_digits=None,
            speech_model=None,
            hints=None,
            language=None,
            interim_speech_results_callback=None,
            interim_speech_results_callback_method=None,
            log=None,
            redirect=None,
            profanity_filter=None,
    ):
        self.add(
            GetInputElement(
                action=action,
                method=method,
                input_type=input_type,
                execution_timeout=execution_timeout,
                digit_end_timeout=digit_end_timeout,
                speech_end_timeout=speech_end_timeout,
                finish_on_key=finish_on_key,
                num_digits=num_digits,
                speech_model=speech_model,
                hints=hints,
                language=language,
                interim_speech_results_callback=interim_speech_results_callback,
                interim_speech_results_callback_method=interim_speech_results_callback_method,
                log=log,
                redirect=redirect,
                profanity_filter=profanity_filter,
            ))
        return self

    def add_hangup(
            self,
            reason=None,
            schedule=None,
    ):
        self.add(HangupElement(
            reason=reason,
            schedule=schedule,
        ))
        return self

    def add_message(
            self,
            content,
            src=None,
            dst=None,
            type=None,
            callback_url=None,
            callback_method=None,
    ):
        self.add(
            MessageElement(
                content=content,
                src=src,
                dst=dst,
                type=type,
                callback_url=callback_url,
                callback_method=callback_method,
            ))
        return self

    def add_play(
            self,
            content,
            loop=None,
    ):
        self.add(PlayElement(
            content=content,
            loop=loop,
        ))
        return self

    def add_pre_answer(self, ):
        self.add(PreAnswerElement())
        return self

    def add_record(
            self,
            action=None,
            method=None,
            file_format=None,
            redirect=None,
            timeout=None,
            max_length=None,
            play_beep=None,
            finish_on_key=None,
            record_session=None,
            start_on_dial_answer=None,
            transcription_type=None,
            transcription_url=None,
            transcription_method=None,
            callback_url=None,
            callback_method=None,
    ):
        self.add(
            RecordElement(
                action=action,
                method=method,
                file_format=file_format,
                redirect=redirect,
                timeout=timeout,
                max_length=max_length,
                play_beep=play_beep,
                finish_on_key=finish_on_key,
                record_session=record_session,
                start_on_dial_answer=start_on_dial_answer,
                transcription_type=transcription_type,
                transcription_url=transcription_url,
                transcription_method=transcription_method,
                callback_url=callback_url,
                callback_method=callback_method,
            ))
        return self

    def add_redirect(
            self,
            content,
            method=None,
    ):
        self.add(RedirectElement(
            content=content,
            method=method,
        ))
        return self

    def add_speak(
            self,
            content,
            voice=None,
            language=None,
            loop=None,
    ):
        self.add(
            SpeakElement(
                content=content,
                voice=voice,
                language=language,
                loop=loop,
            ))
        return self

    def add_wait(
            self,
            length=None,
            silence=None,
            min_silence=None,
            beep=None,
    ):
        self.add(
            WaitElement(
                length=length,
                silence=silence,
                min_silence=min_silence,
                beep=beep,
            ))
        return self

    def add_multi_party_call(self,
                             content,
                             role,
                             max_duration=14400,
                             max_participants=10,
                             wait_music_url=None,
                             wait_music_method="GET",
                             agent_hold_music_url=None,
                             agent_hold_music_method='GET',
                             customer_hold_music_url=None,
                             customer_hold_music_method='GET',
                             recording_callback_url=None,
                             recording_callback_method='GET',
                             status_callback_url=None,
                             status_callback_method='GET',
                             on_exit_action_url=None,
                             on_exit_action_method='POST',
                             record=False,
                             record_file_format='mp3',
                             status_callback_events='mpc-state-changes,participant-state-changes',
                             stay_alone=False,
                             coach_mode=True,
                             mute=False,
                             hold=False,
                             start_mpc_on_enter=True,
                             end_mpc_on_exit=False,
                             relay_dtmf_inputs=False,
                             enter_sound='beep:1',
                             enter_sound_method='GET',
                             exit_sound='beep:2',
                             exit_sound_method='GET'
                             ):
        self.add(
            MultiPartyCallElement(
                content=content,
                role=role,
                max_duration=max_duration,
                max_participants=max_participants,
                wait_music_url=wait_music_url,
                wait_music_method=wait_music_method,
                agent_hold_music_url=agent_hold_music_url,
                agent_hold_music_method=agent_hold_music_method,
                customer_hold_music_url=customer_hold_music_url,
                customer_hold_music_method=customer_hold_music_method,
                record=record,
                record_file_format=record_file_format,
                recording_callback_url=recording_callback_url,
                recording_callback_method=recording_callback_method,
                status_callback_events=status_callback_events,
                status_callback_url=status_callback_url,
                status_callback_method=status_callback_method,
                stay_alone=stay_alone,
                coach_mode=coach_mode,
                mute=mute,
                hold=hold,
                start_mpc_on_enter=start_mpc_on_enter,
                end_mpc_on_exit=end_mpc_on_exit,
                enter_sound=enter_sound,
                enter_sound_method=enter_sound_method,
                exit_sound=exit_sound,
                exit_sound_method=exit_sound_method,
                on_exit_action_url=on_exit_action_url,
                on_exit_action_method=on_exit_action_method,
                relay_dtmf_inputs=relay_dtmf_inputs,
            )
        )
        return self

    def to_dict(self):
        return {}



# -*- coding: utf-8 -*-
"""
Recording class - along with its list class
"""

from plivo.base import (ListResponseObject, PlivoResource,
                        PlivoResourceInterface)
from plivo.resources.accounts import Subaccount
from plivo.utils import is_valid_time_comparison, to_param_dict
from plivo.utils.validators import *


class Recording(PlivoResource):
    _name = 'Recording'
    _identifier_string = 'recording_id'

    def delete(self):
        return self.client.recordings.delete(self.id)


class Recordings(PlivoResourceInterface):
    _resource_type = Recording

    @validate_args(
        subaccount=[optional(is_subaccount())],
        call_uuid=[optional(of_type(six.text_type))],
        limit=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda limit: 0 < limit <= 20, '0 < limit <= 20')))
        ],
        offset=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda offset: 0 <= offset, '0 <= offset')))
        ])
    def list(self,
             subaccount=None,
             call_uuid=None,
             add_time__gt=None,
             add_time__gte=None,
             add_time__lt=None,
             add_time__lte=None,
             add_time=None,
             limit=20,
             offset=0):

        if subaccount:
            if isinstance(subaccount, Subaccount):
                subaccount = subaccount.id

        if add_time__gt and is_valid_time_comparison(add_time__gt):
            add_time__gt = str(add_time__gt)

        if add_time__gte and is_valid_time_comparison(add_time__gte):
            add_time__gte = str(add_time__gte)

        if add_time__lt and is_valid_time_comparison(add_time__lt):
            add_time__lt = str(add_time__lt)

        if add_time__lte and is_valid_time_comparison(add_time__lte):
            add_time__lte = str(add_time__lte)

        if add_time and is_valid_time_comparison(add_time):
            add_time = str(add_time)

        return self.client.request(
            'GET',
            ('Recording', ),
            to_param_dict(self.list, locals()),
            objects_type=Recording,
            response_type=ListResponseObject,
            is_voice_request=True
        )

    @validate_args(recording_id=[of_type(six.text_type)])
    def get(self, recording_id):
        return self.client.request(
            'GET', ('Recording', recording_id), response_type=Recording, is_voice_request=True)

    @validate_args(recording_id=[of_type(six.text_type)])
    def delete(self, recording_id):
        return self.client.request('DELETE', ('Recording', recording_id), is_voice_request=True)
